/*

Author: Sjors Hettinga
Date: 14 December 2010
Hardware: ATmega88, with RFM01 en RFM02 modules
			pinout is defined in rf_pin_config.h and in the corresponding schematic
Description: This file implements the packet layer

	Packets are configures as follows:
	--------------------------------------------------------------------
	|| opt8 | len || opt0-7 || crc1 || data0 || ... || data n || crc2 ||
	--------------------------------------------------------------------
	Byte numbers, starting at 1:
	1 )	From the first byte transmitted, the lower 7 bits indicate the number of data bytes send.
		The MSB is bit 8 in the header option.
	2 ) Header option bits 0 - 7
	3 ) The first CRC8, which checks also if the packet lenght is transmitted correctly
	4 .. 3 + len ) Data bytes, that are the actual payload of the packet
	4 + len ) The second CRC8, checking the data

	A complete packet is transmitted by this function
	Also a callback is implemented, that calls a defined function when a packet has arrived
*/

// AVR lib-C functions
#include <avr/io.h>
#include <avr/interrupt.h>

// custom includes
#include "config.h" // configuration settings
#include "packet.h" // function prototypes for this file
#include "radio.h" // function for the radio layer
#include "xitoa.h" // used for debug on the serial port
#include "crc.h" // CRC function

// receive buffers
volatile packetContents RXpacketBuffers[2];
volatile packetContents* currentRXbuffer;

// working variable for receiving
volatile unsigned char RXcrc;
volatile unsigned int RXpktcounter;

// packet to transmit
volatile unsigned char *TXbuffer;
volatile unsigned char TXlength;
volatile unsigned int TXheaderOpt;

// working variables for transmission
unsigned char TXcrc;
unsigned char TXpktcounter;

// local functions
void readReceivedData(unsigned char data); // process the received byte
unsigned char fillTXbyte(void); // returns the next byte to transmit
void emptyPktReceiver(packetContents* packet); // dummy packter receive function

// function-pointer for the packet receive function
void (*receivePkt)(packetContents*);


// this function inits the packet layer
// it inits
	// the nessasary local variables
	// the radio layer
	// the callback
void initPacket() {
	// init the receive buffers
	RXpacketBuffers[0].nextPacketContents = (void*)&RXpacketBuffers[1];
	RXpacketBuffers[1].nextPacketContents = (void*)&RXpacketBuffers[0];
	currentRXbuffer = &RXpacketBuffers[0];
	
	// init at a dummy function
	receivePkt = emptyPktReceiver;

	// set the function pointers for the driver
	setDataFeeder(fillTXbyte);
	setDataReceiver(readReceivedData);

	// init the driver
	initPHY();
	RXpktcounter = 0;
}

// set the decoder function
void setPktReceiver(void (*myPktReceiver)(packetContents*)){
	receivePkt = myPktReceiver;
}

// empty message decoder function to avoid problemens when not set
void emptyPktReceiver(packetContents* packet){

}

// this function transmits a packet
void sendPacket(unsigned char *buffer, unsigned char len, unsigned int headerOpt) {
	TXbuffer = buffer;
	TXlength = len;
	TXheaderOpt = headerOpt;
	TXpktcounter = 0;

	// start the sender (radio layer)
	startTx();
}

// this fuction returns the n-th byte for transmission
inline char getTXdata(unsigned char counter) {
	return   TXbuffer[counter];
}

// this function gives the next byte to transmit
unsigned char fillTXbyte(void) 
{
	unsigned char TXpktbyte = 0;
	signed char rmCnt;
	switch (TXpktcounter) {
		// this preamble is automatically recognized and processed by the receiver
		// therefore it has to be transmitted at the start of a packet, but is never
		// outputted by the receiver
		case 0: TXpktbyte = 0xAA; break;
		case 1: TXpktbyte = 0xAA; break;
		case 2: TXpktbyte = 0xAA; break;
		case 3: TXpktbyte = 0x2D; break;
		case 4: TXpktbyte = 0xD4; break;

		// set length of data. mark MSB for the type of the packet
		case 5: TXpktbyte = TXlength & 0x7F; // length of the packet
				// 8-th bit of the header options
				TXpktbyte |= (unsigned char)( (TXheaderOpt >> 1) & 0x80 );
				TXcrc = 0; // reset the crc
				calcCRC8(&TXcrc,TXpktbyte);
				break;
		case 6: // transmit the remaining header options
				TXpktbyte = (unsigned char)( TXheaderOpt );
				calcCRC8(&TXcrc,TXpktbyte);
				break;
		case 7: // transmit the crc and reset it
				TXpktbyte = TXcrc;
				TXcrc = 0;
				break;
		case 255:	// idle
				break;
		
		// transmit the data
		default: 
				rmCnt = (signed char)TXpktcounter - (signed char)TXlength;
				switch (rmCnt) {
					case 8: 	TXpktbyte = TXcrc; break; // send crc
					case 9: 	TXpktbyte = 0xAA; break; // append 0xAA to keep things working
					case 10:	xputs(PSTR("Packet send\n")); // shown that a message has been send
								stopTx();
								TXpktcounter = 255; // put in idle mode
								break; // stop the transmitter
					default:
							// just send normal data
							TXpktbyte = getTXdata(TXpktcounter-8);
							calcCRC8(&TXcrc,TXpktbyte);
				}
	}
	TXpktcounter++; // go to the next stage
	return TXpktbyte; // return the data
}

// store the received data
inline void storeRXdata(unsigned char data, unsigned char counter) {
	currentRXbuffer->buffer[counter] = data;
}

// process the data from the receiver
void readReceivedData(unsigned char data){
	switch (RXpktcounter) {
		case 0 : // first byte = length
			currentRXbuffer->length = data & 0x7F;
			currentRXbuffer->headerOpt = ((unsigned int)(data & 0x80)) << 1; // store the header options
			// init de CRC
			RXcrc = 0;
			calcCRC8(&RXcrc,data);
			RXpktcounter++;
			break;
		case 1 : // second byte: additional header info
			currentRXbuffer->headerOpt |= data;
			calcCRC8(&RXcrc,data);
			RXpktcounter++;
			break;
		case 2 : // third byte: checksum header
			if (data != RXcrc) {
				// reset transmission when not correct
				resetRxFIFO(); // reset the receiver to wait for the next preamble (radio layer)
				RXpktcounter = 0;
				xputs(PSTR("Packet error\n"));
			}
			// reset the crc
			RXcrc = 0;
			RXpktcounter++;
			break;
		default :
			if (RXpktcounter == (currentRXbuffer->length + 3)) {
				// pakket fully received
				packetContents* thisPacket = 0;
				if (data == RXcrc){
					// data has been received correctly
					xputs(PSTR("Packet correct\n"));
					// set the pointer
					thisPacket = (packetContents*)currentRXbuffer;
					// switch to the next buffer
					currentRXbuffer = currentRXbuffer->nextPacketContents;
				} else {
					// error in the data
					xputs(PSTR("Packet error\n"));
				}
				// reset the receive to wait for the next preamble (radio layer)
				resetRxFIFO();

				// reset the reveiver
				RXpktcounter = 0;

				// process the data, outside the interrupt!
				sei();
				if (thisPacket != 0) {
					(*receivePkt)(thisPacket);
				}
			} else {
				// store the received data
				storeRXdata(data,RXpktcounter - 3);
				calcCRC8(&RXcrc,data);
				RXpktcounter++;
			}		
	}
}

// function to check if the transmitter is busy
unsigned char isTransmitterBusy(void){
	if (TXpktcounter == 255) {
		return 0;
	} else {
		return 1;
	}
}
