/*
Author: Sjors Hettinga
	based on the examples in datasheets of the RFM01 and RFM02
Date: 30-11-2010
Hardware: ATmega88, met RFM01 en RFM02 modules
Description: This file implements the driver functions for the transmitter and receiver
	The driver has to handle all the harware (SPI, I/O pins en external interrupts)
*/

// AVR lib-C functions
#include <avr/io.h>
#include <avr/interrupt.h>

// include custom functions
#include "config.h"
#include "rf_pin_config.h"
#include "radio.h"
#include "xitoa.h"

// the calculation of the baud rate setting
#define RADIO_BR_SETTING 	10000000UL/(29UL*RF_BAUD_RATE)

// variables used for transmission
volatile unsigned char TXrdbyte;
volatile unsigned int TXrdcounter;

// pointers voor data handling functions
volatile unsigned char (*dataFeederPtr)(void);
volatile void (*dataReceiverPtr)(unsigned char);

// local functions
void RF02_WRT_CMD(unsigned int aCmd); // set registers for the RFM02 module
unsigned int RF01_WRT_CMD(unsigned int aCmd); // set registers for the RFM01 module
inline unsigned char RF01_RDFIFO(void); // read the FIFO for the RFM01 module
inline void sendBit(); // send a bit using the transmittor

// init functions
void RFXX_PORT_INIT(void); // init the I/O pins
void initRadio(); // init the registers of the modules
void initInterrupts(); // init the external interrups

// init the I/O pins
void RFXX_PORT_INIT(void){
	HI_SEL_T();
	HI_SEL_R();
	HI_SDI();
	LOW_SCK();
	SEL_T_OUTPUT();
	SEL_R_OUTPUT();
	SDI_OUTPUT();
	SDO_INPUT();
	INT_T_INPUT();
	INT_R_INPUT();
	SCK_OUTPUT();
	RF02_SEND_ONE();
	RF02_INIT_DATA();//SET DATA OUTPUT
}

// set registers for the RFM02 module
void RF02_WRT_CMD(unsigned int aCmd){
	LOW_SEL_T(); // lower the chip select for the transmitter

	SPSR |= (1<<SPIF); // clear the interrupt flag

	SPDR = aCmd >> 8;
	while ((~SPSR)&(1<<SPIF)) {}; // wait till the SPI transmission has finished
	SPDR = aCmd & 0x00FF;
	while ((~SPSR)&(1<<SPIF)) {}; // wait till the SPI transmission has finished

	HI_SEL_T(); // deselect the transmitter
}

// set (and read from) registers for the RFM02 module
unsigned int RF01_WRT_CMD(unsigned int aCmd){
	unsigned int temp; // is used to store any received data

	LOW_SEL_R(); // lower the chip select for the receiver
	SPSR |= (1<<SPIF); // clear the interrupt flag

	SPDR = aCmd >> 8;
	while ((~SPSR)&(1<<SPIF)) {}; // wait till the SPI transmission has finished
	temp = SPDR;
	temp = temp << 8;
	SPDR = aCmd & 0x00FF;
	while ((~SPSR)&(1<<SPIF)) {}; // wait till the SPI transmission has finished
	temp |= temp;

	HI_SEL_R(); // deselect the receiver
	return temp; // return the received data
}

// read the received data
inline unsigned char RF01_RDFIFO(void){
	unsigned char Result;
	LOW_SEL_R(); // lower the chip select for the receiver
	SPSR |= (1<<SPIF); // clear the interrupt flag
	SPDR = 0;
	while ((~SPSR)&(1<<SPIF)) {}; // wait till the SPI transmission has finished
	SPDR = 0;
	while ((~SPSR)&(1<<SPIF)) {}; // wait till the SPI transmission has finished
	SPDR = 0;
	while ((~SPSR)&(1<<SPIF)) {}; // wait till the SPI transmission has finished
	Result = SPDR;
	HI_SEL_R(); // deselect the receiver
	return Result; // return the resulting data
}

// initialize the radio registers
void initRadio(){
	// changed setting with respect to the data sheets
	RF02_WRT_CMD(0xCC00);
	RF02_WRT_CMD(0x9363);//868BAND,+/-150kHz
	RF02_WRT_CMD(0xA640);//868MHz
	RF02_WRT_CMD(0xB000); // 0dB output power
	RF02_WRT_CMD(0xD280);//PLL settings
	RF02_WRT_CMD(0xC800 + RADIO_BR_SETTING);//19.2kbps
	RF02_WRT_CMD(0xC220);//ENABLE BIT SYNC
	RF02_WRT_CMD(0xC001);//CLOSE ALL

	RF01_WRT_CMD(0x0000);
	RF01_WRT_CMD(0x9182);//868BAND,400kHz
	RF01_WRT_CMD(0xA640);//868MHz
	RF01_WRT_CMD(0xC800 + RADIO_BR_SETTING);//19.2kbps
	RF01_WRT_CMD(0xC683);//AFC setting
	RF01_WRT_CMD(0xC4AC);
	RF01_WRT_CMD(0xC240);//output 1.66MHz
	RF01_WRT_CMD(0xC080);
	RF01_WRT_CMD(0xCE88);//use FIFO
	RF01_WRT_CMD(0xCE8B);
	RF01_WRT_CMD(0xC081);

}

// initialize the interrupt
void initInterrupts(){
    // falling edge, as from the datasheet derived
	// this is different from the example code in the datasheet!
	EICRA = 2 << (2 * TX_INT_NR);
	// as long as the interrupts is low
	EICRA |= 0 << (2 * RX_INT_NR);

	EIFR = 0x03; // clear interrupt flags
	EIMSK = 0x03; // enable both interrupts
}

// transmitter interrupt
ISR(TX_INT_VEC) {
	sendBit(); // send a single bit
}

// receiver interrupt
ISR(RX_INT_VEC) {
	unsigned char data = RF01_RDFIFO(); // read received data
	sei(); // stop blocking interrupt
	// handle data without having to wait
	(*dataReceiverPtr)(data);
}

// send a single bit
inline void sendBit(){
	// send the right bit
	if(TXrdbyte & 0x80){
		RF02_SEND_ONE();
	}else{
		RF02_SEND_ZERO();
	}
	TXrdbyte = TXrdbyte << 1;

	// check if a new byte is required
	TXrdcounter++;
	if (!(TXrdcounter & 0x0007)) {
		// fill the buffer again
		TXrdbyte = (*dataFeederPtr)();
	}
}

// ---- external visible funtions

// config the physical layer
void initPHY(void) {
	// init ports
	RFXX_PORT_INIT();

	// init SPI bus
	SPCR = (1<<SPE) | (1<<MSTR) | (1<<SPR0); // master, klok/16

	// init radio
	initRadio();
	// init interrupts
	initInterrupts();
	
	stopTx();
	resetRxFIFO();
}

// start transmission by configuring the transmitter for transmission
void startTx(void) {
	// feed new data
	TXrdcounter = 0;
	TXrdbyte = (*dataFeederPtr)(); // feed the next byte
	// start the sender
	RF02_WRT_CMD(0xC039);//START TX
}

// stop the transmitter
void stopTx(void) {
	RF02_WRT_CMD(0xC001);
}

// reset FIFO for next frame recognition
void resetRxFIFO() {
	RF01_WRT_CMD(0xCE88);
	RF01_WRT_CMD(0xCE8B);
}	

// configs the data handlers
// set the function that is calles when the transmitter needs a new byte
void setDataFeeder(unsigned char (*dataFeed)(void) ) {
	dataFeederPtr = (volatile unsigned char (*)(void))dataFeed;
}

// set the function to call when the receiver has a new byte
void setDataReceiver(void (*dataReceiver)(unsigned char) ) {
	dataReceiverPtr = (volatile void (*)(unsigned char))dataReceiver;
}
